import numpy as np
from typing import Union, Callable
from collections.abc import Sequence


class ScatteringFunction:
    """A class to represent a scattering function.

    Abstract base class for scattering functions that can be
    evaluated at a given wavevector (kx, ky, kz). Subclasses should
    implement the `__call__` method.
    """
    def __init__(self, params):
        self.params = params
    
    def __call__(self, kx, ky, kz, **kwargs):
        """Evaluate the scattering function at the given wavevector.

        Parameters
        ----------
        kx, ky, kz : float
            The components of the wavevector in Cartesian coordinates.

        Returns
        -------
        float
            The value of the scattering function at
            the given wavevector.
        """
        raise NotImplementedError("Subclasses should implement this method.")


class IsotropicScattering(ScatteringFunction):
    """A class for isotropic scattering.

    This class represents a scattering function that is constant
    everywhere, defined by the parameter `gamma_0`.
    """
    def __call__(self, kx, ky, kz, **kwargs):
        return self.params['gamma_0']


class AzimuthalScattering(ScatteringFunction):
    """A class for azimuthal scattering.

    This class represents a scattering function that depends on the
    azimuthal angle of the wavevector in the x-y plane, defined by
    parameters `gamma_k`, `power`, `trig`, and `sym`. The scattering
    function is defined as `gamma_k * abs(trig(sym * phi))^power`,
    where `phi` is the angle of the projection of the wavevector k in
    the x-y plane with the x axis, and `trig` is a trigonometric
    function (cos, sin, tan, or cot) depending on the `trig` parameter.
    """
    def __init__(self, params):
        super().__init__(params)
        trig_funcs = {'cos': np.cos, 'sin': np.sin, 'tan': np.tan, 'cot': _cot}
        self.trig_func = trig_funcs[params['trig']]
        if 'sym' not in params:
            self.params['sym'] = 1
    def __call__(self, kx, ky, kz, **kwargs):
        phi = np.arctan2(ky, kx)
        return (self.params['gamma_k'] * np.abs(self.trig_func(
            self.params['sym']*phi)) ** self.params['power'])


class ScatteringSum(ScatteringFunction):
    """A class for summing multiple scattering functions.

    This class represents a scattering function that is the sum of
    multiple scattering functions, each defined by its own parameters.
    It allows for combining different scattering models into a single
    function.
    """
    def __init__(self, scattering_functions):
        super().__init__(params={})
        self.scattering_functions = scattering_functions
    
    def __call__(self, kx, ky, kz, **kwargs):
        return sum(s(kx, ky, kz, **kwargs) for s in self.scattering_functions)


def build_scattering_function(
        scattering_params: dict[str, Union[float, Sequence[float]]],
        scattering_models: Sequence[str] = ['isotropic']):
    """Build a scattering function from the given parameters.

    Supported scattering models include:

    * ``'isotropic'``: Constant ``gamma_0`` everywhere
    * | ``'cos'``: ``gamma_k * abs(cos(sym * phi))^power`` where `phi`
      | is the angle of the projection of the wavevector k in the x-y
      | plane with the x axis. The rest are parameters of the model.
    * | ``'sin'``, ``'tan'``, and ``'cot'``: Same as ``'cos'`` but
      | using different trigonometric functions.
    * | ``'cos[n]phi'``: Where [n] is some integer, e.g. ``'cos2phi'``.
      | Alias for ``'cos'`` with sym being set to the integer in [n].
    * | ``'sin[n]phi'``, ``'tan[n]phi'``, and ``'cot[n]phi'``: Same as
      | ``'cos[n]phi'`` but using different trigonometric functions.

    Parameters
    ----------
    scattering_params : dict[str, float or Sequence[float]]
        Dictionary mapping the names of the parameters to their value
        in each scattering model. If the value is a single number, it
        is assumed to be the parameter for all models. 
    scattering_models : Sequence['str'], optional
        The type of scattering model to use.
    
    Returns
    -------
    Callable
        A class that implements the scattering function and is callable
        on (kx, ky, kz).
    """

    scattering_functions = []
    for i, model in enumerate(scattering_models):
        if model == 'isotropic':
            scattering_functions.append(IsotropicScattering(
                {'gamma_0': _get_param(scattering_params, 'gamma_0', i)}))
        elif model.startswith('cos') or model.startswith('sin'):
            params = _get_params(scattering_params, ['gamma_k', 'power'], i)
            params['trig'] = model[:3]
            if len(model) > 3:
                params['sym'] = int(model[3:-3])
            scattering_functions.append(AzimuthalScattering(params))
    return ScatteringSum(scattering_functions)


def _get_param(scattering_params, key, idx):
        if isinstance(scattering_params[key], (int, float)):
            return scattering_params[key]
        else:
            return scattering_params[key][idx]


def _get_params(scattering_params, keys, idx):
    return {key: _get_param(scattering_params, key, idx) for key in keys}


def _cot(x):
    """Compute the cotangent of x. Defined to avoid lambda function"""
    return 1.0 / np.tan(x) if x != 0 else np.inf



from .bandstructure import BandStructure

import numpy as np
import scipy.sparse

from typing import Callable, Union
from collections.abc import Sequence

from scipy.constants import e, hbar, angstrom


class Conductivity:
    """
    Calculates the conductivity of a material solving the Boltzmann
    transport equation using a finite element method (FEM).

    Parameters
    ----------
    band : BandStructure
        The class holding band structure information of the material.
    field : Sequence[float]
        The magnetic field in the x, y, and z directions in units of
        Tesla.
    scattering_rate : Callable or float or None
        The (out-)scattering rate as a function of kx, ky, and kz, in
        units of THz. Can also be a constant value instead of a
        function. If None, it will be calculated from the scattering
        kernel.
    scattering_kernel : Callable or None
        The scattering kernel as a function of a pair of coordinates
        ``(kx, ky, kz)`` and ``(kx', ky', kz')``, in units of angstrom
        THz. All coordinates are given to the function in order, so the
        function signature would be ``C(kx, ky, kz, kx', ky', kz')``.
        If None, the scattering rate should be specified instead.
    scattering_params : dict[str, float or Sequence[float]], optional
        Extra parameters passed to the scattering kernel or the
        scattering rate function.
    frequency : float
        The frequency of the applied field in units of THz.
    correct_curvature : bool, optional
        If True, correct for the curvature of the Fermi surface.
    solver : Callable, optional
        The solver used to solve the linear system. Takes the (sparse)
        matrix as the first argument and the right-hand side as the
        second argument. When using a custom solver, keep in mind that
        the right-hand side might not be a vector. So, solvers that
        only work with vectors need to be adapted to solve each column
        of the right-hand side separately.
    
    Attributes
    ----------
    band : BandStructure
        The class holding band structure information of the material.
    field : Sequence[float] or None
        The magnetic field in the x, y, and z directions in units of
        Tesla.
    scattering_rate : Callable or float or Sequence[float] or None
        The (out-)scattering rate as a function of kx, ky, and kz. Can
        also be a constant value instead of a function. If initialized
        as None, it will be calculated from the scattering kernel upon
        the next calculation.
    scattering_kernel : Callable or None
        The scattering kernel as a function of a pair of coordinates
        ``(kx, ky, kz)`` and ``(kx', ky', kz')`` in units of angstrom,
        which returns values in units of angstrom THz. All coordinates
        are given to the function in order, so the function signature
        would be ``C(kx, ky, kz, kx', ky', kz')``. If None, the
        scattering rate should be specified instead. The out-scattering
        rate will be calculated from the scattering kernel if the
        scattering rate is not provided.
    scattering_params : dict[str, float or Sequence[float]], optional
        Extra parameters passed to the scattering kernel or the
        scattering rate function. If, for instance, you also want the
        scattering to depend on the velocities, you can set them here.
    frequency : float
        The frequency of the applied field in units of THz.
        If non-zero, the conductivity output will be complex.
    sigma : numpy.ndarray
        The conductivity tensor, which is a 3 by 3 matrix. Can be
        calculated using the ``solve`` method. Elements that are not
        calculated yet are set to zero.
    correct_curvature : bool
        Whether to correct for the curvature of the Fermi surface.
    solver : Callable
        The solver used to solve the linear system.
    """
    def __init__(
            self, band: BandStructure, field: Sequence[float] = np.zeros(3),
            scattering_rate: Union[Callable, float, None] = None,
            scattering_kernel: Union[Callable, None] = None,
            scattering_params: dict[str, Union[float, Sequence[float]]] = {},
            frequency: float = 0.0, correct_curvature: bool = True,
            solver: Callable = scipy.sparse.linalg.spsolve, **kwargs):
        self.solver = solver
        self.correct_curvature = correct_curvature
        # avoid triggering setattr in the constructor
        super().__setattr__('band', band)
        super().__setattr__('scattering_rate', scattering_rate)
        super().__setattr__('scattering_kernel', scattering_kernel)
        super().__setattr__('scattering_params', scattering_params)
        super().__setattr__('frequency', frequency)
        super().__setattr__('field', np.array(field))
        self._field_magnitude = np.linalg.norm(field)
        if self._field_magnitude != 0:
            self._field_direction = field / self._field_magnitude
        else:
            self._field_direction = np.zeros(3)
        self.sigma = np.zeros((3, 3))
        self._saved_solutions = [None, None, None]
        self._velocities = None
        self._vmags = None
        self._vhats = None
        self._vhat_projections = None
        self._jacobians = None
        self._jacobian_sums = None
        self._derivative_components = None
        self._derivatives = None
        self._scattering_invlen = None
        self._out_scattering = None
        self._derivative_term = None
        self._differential_operator = None
        self._are_elements_saved = False
        self._is_scattering_saved = False
        self._saved_solutions = [None, None, None]

    def __setattr__(self, name, value):
        if name == 'band':
            self.erase_memory()
        if name in ['frequency', 'scattering_rate',
                    'scattering_kernel', 'scattering_params']:
            self.erase_memory(elements=False, scattering=True,
                              derivative=False)
        if name == 'field' and value is not None:
            self.set_field(value)
        super().__setattr__(name, value)
    
    def set_field(self, field):
        field = np.array(field)
        new_magnitude = np.linalg.norm(field)
        if new_magnitude != 0:
            new_direction = field / new_magnitude
        else:
            new_direction = np.zeros(3)
        if np.all(self._field_direction == new_direction):
            if self._derivative_term is not None:
                self._derivative_term *= \
                    new_magnitude / self._field_magnitude
            if self._differential_operator is not None:
                self._differential_operator = \
                    self._out_scattering - e/hbar*self._derivative_term
                self._saved_solutions = [None, None, None]
        else:
            self.erase_memory(elements=False, scattering=False,
                              derivative=True)
        self._field_magnitude = new_magnitude
        self._field_direction = new_direction
        super().__setattr__('field', field)

    def calculate(self, i: Union[Sequence[int], int, None] = None,
                  j: Union[Sequence[int], int, None] = None
                  ) -> Union[np.ndarray, float]:
        """Calculate the conductivity tensor.

        Parameters
        ----------
        i : Sequence[int] or int or None, optional
            The index of the first component (row) of the conductivity
            tensor. If None (default), all components are calculated.
        j : Sequence[int] or int or None, optional
            The index of the second component (column) of the
            conductivity tensor. If None (default), all components
            are calculated.
        Returns
        -------
        numpy.ndarray or float
            The conductivity tensor component(s) as an i by j matrix.
        """
        if not self._are_elements_saved:
            self._build_elements()
        if self._differential_operator is None:
            self._build_differential_operator()
        
        i, j, j_calc = self._get_calculation_indices(i, j)
        # (A^{-1})^{ij} (v_b)_j
        linear_solution = self.solver(self._differential_operator,
                                      self._vhat_projections[:, j_calc])
        if len(linear_solution.shape) == 1:
            linear_solution = linear_solution[:, None]
        # reuse previously calculated solutions
        for col in j:
            if col in j_calc:
                # save solution for potential reuse
                self._saved_solutions[col] = \
                    linear_solution[:, j_calc.index(col)]
            else:
                linear_solution = np.insert(
                    linear_solution, col, self._saved_solutions[col], axis=1)
        # (v_a)_i (A^{-1} v_b)^i
        sigma_result = self._vhat_projections[:, i].T @ linear_solution
        sigma_result *= e**2 / (4 * np.pi**3 * hbar)

        for idx_row, row in enumerate(i):
            for idx_col, col in enumerate(j):
                self.sigma[row, col] = sigma_result[idx_row, idx_col]
        return sigma_result

    def erase_memory(self, elements: bool = True, scattering: bool = True,
                     derivative: bool = True):
        """Erase saved calculations to free memory.

        This class saves already calculated values for the FEM elements
        and matrices to avoid recalculating them every time a new
        element of the conductivity tensor is calculated or a new field
        is applied. This method is provided to erase those values when
        no new calculations are needed and the memory can be freed.

        Parameters
        ----------
        elements : bool, optional
            If True, erase the quantities for each element, like the
            lengths and velocities.
        scattering : bool, optional
            If True, erase the out-scattering and in-scattering terms.
        derivative : bool, optional
            If True, erase the derivative term.
        """
        if elements:
            self._velocities = None
            self._vmags = None
            self._vhats = None
            self._jacobians = None
            self._jacobian_sums = None
            self._derivative_components = None
            self._derivatives = None
            self._vhat_projections = None
            self._are_elements_saved = False
        if scattering:
            self._scattering_invlen = None
            self._out_scattering = None
            self._is_scattering_saved = False
        if derivative:
            self._derivative_term = None
        self._differential_operator = None
        self._saved_solutions = [None, None, None]

    def _get_calculation_indices(self, i, j):
        if i is None:
            i = range(3)
        elif isinstance(i, int):
            i = [i]
        if j is None:
            j = range(3)
        elif isinstance(j, int):
            j = [j]
        j_calc = []
        for col in j:
            if self._saved_solutions[col] is None:
                j_calc.append(col)
        return i, j, j_calc

    def _build_elements(self):
        """
        Build the arrays corresponding to the discretization of the
        band structure.
        """
        self._velocities = np.column_stack(self.band.velocity_func(
            self.band.kpoints[:, 0], self.band.kpoints[:, 1],
            self.band.kpoints[:, 2]))
        self._vmags = np.linalg.norm(self._velocities, axis=1)
        self._vhats = self._velocities / self._vmags[:, None]

        triangle_points = self.band.kpoints[self.band.kfaces] / angstrom
        if self.correct_curvature:
            triangle_points = self._curvature_correct_points(triangle_points)
    
        self._calculate_jacobian_sums(triangle_points)
        self._calculate_derivative_sums(triangle_points)
        self._calculate_velocity_projections()
        self._are_elements_saved = True
    
    def _curvature_correct_points(self, triangle_points):
        """Take the curvature into account for the areas and lengths"""
        kcenters = np.mean(triangle_points, axis=1) * angstrom
        kcenters_tangent = kcenters.copy()
        for _ in range(2):
            kcenters_tangent = self.band._apply_newton_correction(
                kcenters_tangent)
        center_diff = (kcenters_tangent - kcenters) / angstrom
        projected_diff = np.linalg.norm(center_diff, axis=-1)

        point_normals = self._vhats[self.band.kfaces]
        # Turn off warnings for division by zero
        with np.errstate(divide='ignore', invalid='ignore'):
            # cosine = nhat.cdiff / |cdiff||nhat| and |nhat| = 1
            cosines = np.einsum('ijk,ik->ij', point_normals, center_diff
                                ) / projected_diff[:, None]
            # Handle division by zero
            np.nan_to_num(cosines, copy=False, nan=1.0)
        diff = projected_diff[:, None] / cosines
        return triangle_points + point_normals*diff[:, :, None]

    def _calculate_jacobian_sums(self, triangle_points):
        """Calculate the Jacobian sums for each point and point pair."""
        self._jacobians = np.linalg.norm(
                np.cross(triangle_points[:, 1] - triangle_points[:, 0],
                         triangle_points[:, 2] - triangle_points[:, 0]),
                axis=-1)
        # build diagonal ordered matrices of the jacobian sums
        n = len(self.band.kpoints)
        i_idx = self.band.kfaces[:, 0]
        j_idx = self.band.kfaces[:, 1]
        k_idx = self.band.kfaces[:, 2]
        rows = np.concatenate((i_idx, j_idx, k_idx, i_idx, i_idx,
                                j_idx, j_idx, k_idx, k_idx))
        cols = np.concatenate((i_idx, j_idx, k_idx, j_idx, k_idx,
                                i_idx, k_idx, i_idx, j_idx))
        self._jacobian_sums = scipy.sparse.csr_array(
            (np.tile(self._jacobians, 9), (rows, cols)), shape=(n, n))
        self._jacobian_diagonal = np.zeros(n)
        np.add.at(self._jacobian_diagonal, i_idx, self._jacobians)
        np.add.at(self._jacobian_diagonal, j_idx, self._jacobians)
        np.add.at(self._jacobian_diagonal, k_idx, self._jacobians)

    def _calculate_derivative_sums(self, triangle_points):
        """
        Calculate the field-independent part of the derivative term.
        """
        self._derivative_components = (
            triangle_points - np.roll(triangle_points, -2, axis=1))
        i_idx = self.band.kfaces
        j_idx = np.roll(self.band.kfaces, -1, axis=1)
        k_idx = np.roll(self.band.kfaces, -2, axis=1)
        rows = np.concatenate((i_idx.flat, k_idx.flat))
        cols = np.tile(j_idx.flat, 2)
        self._derivatives = [
            scipy.sparse.csc_array((
            np.tile(component.flat, 2), (rows, cols))) for component in
            self._derivative_components.transpose(2, 0, 1)]
        if self.band.periodic:
            self._derivatives = [
                (self.band.periodic_projector @ derivative
                 @ self.band.periodic_projector.T).tocsc()
                for derivative in self._derivatives]

    def _calculate_velocity_projections(self):
        self._vhat_projections = self._jacobian_sums @ self._vhats / 24
        self._vhat_projections += (
            self._jacobian_diagonal[:, None] * self._vhats / 24)
        if self.band.periodic:
            self._vhat_projections = \
                self.band.periodic_projector @ self._vhat_projections

    def _build_differential_operator(self):
        """
        Build the differential operator from the elements of the
        band structure and the conductivity information.
        """
        if not self._is_scattering_saved:
            self._discretize_scattering()
            self._build_out_scattering_matrix()
            # TODO: calculate the in-scattering matrix
            self._is_scattering_saved = True
        if self._derivative_term is None:
            self._derivative_term = sum(
                Bi / 6 * Di for Bi, Di in zip(self.field, self._derivatives))
        self._differential_operator = (
            self._out_scattering - e/hbar*self._derivative_term)
            # - self._in_scattering_term when implemented

    def _discretize_scattering(self):
        """
        Discretize the scattering rate and the scattering kernel
        for each element.
        """
        if self.scattering_rate is None:
            if self.scattering_kernel is None:
                raise ValueError(
                    "Either scattering_rate or scattering_kernel must be set.")
            else:
                self._calculate_out_scattering_from_kernel()

        if isinstance(self.scattering_rate, Callable):
            scattering = self.scattering_rate(
                self.band.kpoints[:, 0], self.band.kpoints[:, 1],
                self.band.kpoints[:, 2], **self.scattering_params)
        else:
            scattering = self.scattering_rate
        # separate the optical conductivity case to avoid making
        # the number complex when it is not needed
        if self.frequency == 0.0:
            self._scattering_invlen = 1e12 * scattering / self._vmags
        else:
            self._scattering_invlen = \
                1e12 * (scattering - 2j*np.pi*self.frequency) / self._vmags
        # scattering_invlen is the inverse scattering length gamma
        # TODO: discretize the scattering kernel

    def _calculate_out_scattering_from_kernel(self):
        """
        Calculate the scattering rate by integrating over
        the scattering kernel.
        """
        # TODO: implement this
        pass

    def _build_out_scattering_matrix(self):
        """Calculate the out-scattering matrix (Gamma)"""
        self._out_scattering = (
            # alpha_ij * gamma^i
            (self._jacobian_sums * self._scattering_invlen[:, None]).tocsc()
            # alpha_ij * gamma^j
            + (self._jacobian_sums * self._scattering_invlen[None, :]).tocsc()
            # sum_k alpha_ik * gamma^k
            + scipy.sparse.diags_array(
                self._jacobian_sums @ self._scattering_invlen, format='csc')
            ) / 60
        # alpha(i,j,k) * gamma^k / 120
        i_idx = self.band.kfaces[:, 0]
        j_idx = self.band.kfaces[:, 1]
        k_idx = self.band.kfaces[:, 2]
        n = len(self.band.kpoints)
        rows = np.concatenate((i_idx, i_idx, j_idx, j_idx, k_idx, k_idx))
        cols = np.concatenate((j_idx, k_idx, i_idx, k_idx, i_idx, j_idx))
        data = np.concatenate((
            self._jacobians * self._scattering_invlen[k_idx],
            self._jacobians * self._scattering_invlen[j_idx],
            self._jacobians * self._scattering_invlen[k_idx],
            self._jacobians * self._scattering_invlen[i_idx],
            self._jacobians * self._scattering_invlen[j_idx],
            self._jacobians * self._scattering_invlen[i_idx])) / 120
        self._out_scattering += scipy.sparse.csc_array(
            (data, (rows, cols)), shape=(n, n))
        if self.band.periodic:
            self._out_scattering = (
                self.band.periodic_projector @ self._out_scattering
                @ self.band.periodic_projector.T).tocsc()



import unittest
from elecboltz.fit import _extract_flat_keys, _build_params_from_flat

class TestFlatDictionary(unittest.TestCase):
    def test_extract_flat_keys(self):
        params = {
            'energy_scale': 100.0,
            'scattering_rate': 1.0,
            'band_params': {
                'a': 2.0,
                'b': 3.0,
                'c': 5.0,
            },
            'scattering_models': ['isotropic', 'cos2phi'],
            'scattering_params': {
                'power': [6, 7],
                'gamma_0': [8.1, 9.2],
                'gamma_k': [52.1, 63.4]
            }
        }
        expected_keys = [
            'energy_scale', 
            'scattering_rate', 
            'band_params.a', 
            'band_params.b', 
            'band_params.c',
            'scattering_models.0',
            'scattering_models.1',
            'scattering_params.power.0', 
            'scattering_params.power.1',
            'scattering_params.gamma_0.0',
            'scattering_params.gamma_0.1',
            'scattering_params.gamma_k.0',
            'scattering_params.gamma_k.1'
        ]
        self.assertEqual(
            _extract_flat_keys(params, bounds=False), expected_keys,
            "Extracted flat keys do not match expected keys "
            "with bounds=False.")

        params = {
            'energy_scale': (100.0, 200.0),
            'scattering_rate': (0.1, 10.0),
            'band_params': {
                'a': (0.1, 1.0),
                'b': (0.1, 1.0),
                'c': (0.1, 1.0),
            },
            'scattering_params': {
                'power': [(1, 20), (3, 15)],
                'gamma_0': [(1.0, 10.0), (1.0, 10.0)],
                'gamma_k': [(1.0, 10.0), (1.0, 10.0)],
            }
        }
        expected_keys = [
            'energy_scale', 
            'scattering_rate', 
            'band_params.a', 
            'band_params.b', 
            'band_params.c',
            'scattering_params.power.0', 
            'scattering_params.power.1',
            'scattering_params.gamma_0.0',
            'scattering_params.gamma_0.1',
            'scattering_params.gamma_k.0',
            'scattering_params.gamma_k.1'
        ]
        self.assertEqual(
            _extract_flat_keys(params, bounds=True), expected_keys,
            "Extracted flat keys do not match expected keys "
            "with bounds=True.")

    def test_build_params_from_flat(self):
        keys = [
            'energy_scale', 
            'scattering_rate', 
            'band_params.a', 
            'band_params.b', 
            'band_params.c', 
            'scattering_params.nu.0', 
            'scattering_params.nu.1', 
            'scattering_params.gamma.0', 
            'scattering_params.gamma.1'
        ]
        values = [100.0, 1.0, 0.1, 0.2, 0.3, 4.0, 5.0, 6.0, 7.0]
        expected_params = {
            'energy_scale': 100.0,
            'scattering_rate': 1.0,
            'band_params': {
                'a': 0.1,
                'b': 0.2,
                'c': 0.3
            },
            'scattering_params': {
                'nu': [4.0, 5.0],
                'gamma': [6.0, 7.0]
            }
        }
        self.assertEqual(
            _build_params_from_flat(keys, values), expected_params,
            "Dictionary built from flat keys"
            "do not match expected parameters.")


if __name__ == '__main__':
    unittest.main()



from .bandstructure import BandStructure
from .conductivity import Conductivity
from .params import easy_params

import numpy as np
from scipy.optimize import differential_evolution
import json
from datetime import datetime
from time import time
from copy import deepcopy
from multiprocessing import cpu_count
from pathlib import Path
from pprint import pformat
from typing import Callable
from collections.abc import Sequence, Collection, Mapping


class FittingRoutine:
    """This convenience class automatically sets up a fitting routine.

    Stores the necessary data and generates the functions used in
    This includes handling the residual computation, parameter update,
    logging the fitting progress, and saving the results to a file.

    Parameters
    ----------
    init_params : Mapping
        Initial parameters for the fitting routine, and also other
        parameters for initiallizing the classes. This is passed through
        ``easy_params`` to the ``BandStructure`` and ``Conductivity``.
    save_path : str, optional
        The directory where the fitting logs will be saved. If not
        provided, logs will not be saved.
    save_label : str, optional
        Label of the results.
    update_keys : Collection[str], optional
        The "flattened" keys of the parameters that will be updated
        during the fitting. See ``extract_keys`` for more information.
        Used for updating the ``params`` attribute and logging. If not
        provided, ``params`` will not be updated and the parameter names
        will not be mentioned in the log.
    print_log : bool, optional
        If True, the fitting progress will be printed to the console.
    
    Attributes
    ----------
    params : Mapping
        The current parameters dictionary.
    save_path : Path
        The directory where the fitting logs will be saved.
    save_label : Path
        The label of the results.
    update_keys : Collection[str]
        The "flattened" keys of the parameters that will be updated
        during the fitting. This is used for updating the `params`
        attribute and logging.
    iteration : int
        The current iteration number.
    last_time : float
        The timestamp of the last fitting iteration.
    total_time : float
        The total time spent on the fitting routine.
    """
    def __init__(self, init_params: Mapping, save_path: str = None,
                 save_label: str = "fit", update_keys: Collection[str] = None,
                 print_log: bool = True):
        self.init_params = init_params
        self.save_path = save_path
        self.save_label = save_label
        self.update_keys = update_keys
        self.print_log = print_log
        self.iteration = 0
        self.last_time = time()
        self.total_time = 0.0

    def residual(
            self, param_values: Sequence, param_keys: Sequence[str],
            x_data: Mapping[str, Sequence], y_data: Mapping[str, Sequence],
            x_shift: Mapping = None, x_normalize: Mapping = None,
            y_shift: Mapping = None, y_normalize: Mapping = None,
            cond_obj: Conductivity = None, loss: Callable =
                lambda y_fit, y_data: np.mean(np.abs(y_fit - y_data))):
        """Compute the residual for the given parameters and data.

        Parameters
        ----------
        param_values : Sequence
            The values of the parameters to update.
        param_keys : Sequence[str]
            The keys of the parameters to update.
        x_data : Mapping[str, Sequence]
            The independent variable data (e.g. field). The name of the
            variable is mapped to the data, e.g.
            ``{'field': [0, 1, 2]}``.
        y_data : Mapping[str, Sequence]
            The dependent variable data (e.g. conductivity). The name
            of the variable is mapped to the data, e.g.
            ``{'sigma_xx': [1.1, 2.4, 3.8]}``.
        x_shift : Mapping, optional
            If provided, the y values will be shifted by the y value at
            this x point.
        x_normalize : Mapping, optional
            If provided, the y values will be normalized by the y value
            at this x point.  Note that shifts are applied before
            normalization.
        y_shift : float, optional
            The y values will be normalized to this value
            (if ``x_normalize`` is provided).
        y_normalize : float, optional
            The y values will be shifted to this value (if ``x_shift``
            is provided).
        squared : bool, optional
            If True, the residual is computed as the mean squared error.
            If False, it returns the mean absolute difference.
        cond_obj : Conductivity, optional
            If provided, the fitter will try to salvage the calculations
            already done from that object.
        """
        cond = self._build_obj(param_values, param_keys, cond_obj)
        name, y_label_i, y_label_j = self._get_label_indices(y_data.keys())

        y_fit = {label: np.zeros_like(y) for label, y in y_data.items()}
        for i in range(len(list(x_data.values())[0])):
            x = {label: x[i] for label, x in x_data.items()}
            y = _get_y(cond, x, y_data, name, y_label_i, y_label_j)
            for label in y_fit:
                y_fit[label][i] = y[label]
        if x_shift is not None:
            y0 = _get_y(cond, x_shift, y_data, name, y_label_i, y_label_j)
            for label in y_fit:
                y_fit[label] -= y0[label]
            if y_shift is not None:
                for label in y_fit:
                    y_fit[label] += y_shift[label]
        if x_normalize is not None:
            y0 = _get_y(cond, x_normalize, y_data, name, y_label_i, y_label_j)
            for label in y_fit:
                y_fit[label] /= y0[label]
            if y_normalize is not None:
                for label in y_fit:
                    y_fit[label] *= y_normalize[label]

        y_fit = np.concatenate(list(y_fit.values()))
        y_data = np.concatenate(list(y_data.values()))
        return loss(y_fit, y_data)

    def log(self, param_values, convergence: float = None):
        """Log the current fitting iteration and parameters.
        
        Parameters
        ----------
        param_values : Sequence
            The current values of the parameters.
        convergence : float, optional
            The convergence value of the fitting routine. 0.0 means
            no convergence, 1.0 means perfect convergence.
            If not provided, it will not be logged.
        """
        self.iteration += 1
        now = time()
        iter_time = now - self.last_time
        self.last_time = now
        self.total_time += iter_time

        log_message = f"Iteration {self.iteration}\n" \
                      f"----------{'-' * len(str(self.iteration))}\n" \
                      f"Best Parameters:\n"
        if self.update_keys is not None:
            update_params = _build_params_from_flat(
                self.update_keys, param_values)
            log_message += pformat(update_params) + "\n"
        else:
            log_message += pformat(param_values) + "\n"
        if convergence is not None:
            log_message += f"Convergence: {convergence:.5f}\n\n"
        log_message += f"Iteration Runtime: {iter_time:.3f} seconds\n"
        minutes, seconds = divmod(self.total_time, 60)
        hours, minutes = divmod(minutes, 60)
        log_message += f"Total Runtime: "
        if hours > 0:
            log_message += f"{int(hours)} hours "
        if minutes > 0:
            log_message += f"{int(minutes)} minutes "
        log_message += f"{seconds:.1f} seconds\n"

        if self.print_log:
            print(log_message)
        if self.save_path is not None:
            path = Path(self.save_path) / f"{self.save_label}.log"
            mode = 'w' if self.iteration == 1 else 'a'
            with open(path, mode) as log_file:
                log_file.write(log_message)

    def _build_obj(self, param_values: Sequence, param_keys: Sequence[str],
                   cond: Conductivity = None) -> Conductivity:
        """
        Build the conductivity object with the given parameters.
        """
        if cond is None:
            band = BandStructure(**easy_params(self.init_params))
            cond = Conductivity(band, **easy_params(self.init_params))
            update_band = True
        else:
            cond = deepcopy(cond)
            band = cond.band
            update_band = False
        params = deepcopy(self.init_params)
        for key, value in zip(param_keys, param_values):
            _update_flat_value(params, key, value)
        new_params = easy_params(params)
        for key, value in new_params.items():
            if key == 'band_params':
                for band_key, band_value in value.items():
                    if band.band_params[band_key] != band_value:
                        update_band = True
                        band.band_params[band_key] = band_value
            if hasattr(band, key):
                if np.any(getattr(band, key) != value):
                    update_band = True
                    setattr(band, key, value)
            if hasattr(cond, key):
                if np.any(getattr(cond, key) != value):
                    setattr(cond, key, value)
        if update_band:
            band.discretize()
            cond.band = band
        return cond

    def _get_label_indices(self, labels: Collection[str]):
        """Extract the names and indices from y_data keys."""
        name, i, j = {}, {}, {}
        for label in labels:
            name[label], index_labels = label.split("_")
            i[label] = {'x': 0, 'y': 1, 'z': 2}[index_labels[0]]
            j[label] = {'x': 0, 'y': 1, 'z': 2}[index_labels[1]]
        return name, i, j


def fit_model(x_data: Mapping[str, Sequence], y_data: Mapping[str, Sequence],
              init_params: Mapping, bounds: Mapping,
              x_shift: Mapping = None, x_normalize: Mapping = None,
              save_path: str = None, save_label: str = None,
              worker_percentage: float = 0.0, **kwargs):
    """Convenience function to set up and run a fitting routine.

    This uses ``scipy.optimize.differential_evolution`` to perform a
    global fit. The optimizer is hard-coded, because the callback
    functions are highly specific to each optimizer.
    Saves the results to the specified path.

    Parameters
    ----------
    x_data : Mapping[str, Sequence]
        The independent variable data (e.g. field). The name of the
        variable is mapped to the data, e.g. ``{'field': [0, 1, 2]}``.
    y_data : Mapping[str, Sequence]
        The dependent variable data (e.g. conductivity). The name of
        the variable is mapped to the data, e.g.
        ``{'sigma_xx': [1.1, 2.4, 3.8]}``. The name of each variable
        must start with "sigma" or "rho" (for conductivity or
        resistivity, respectively), and you can add a suffix to specify
        the component (e.g. ``"sigma_xx"``, ``"rho_xy"``).
    init_params : Mapping
        Initial parameters for the fitting routine, and also other
        parameters for initiallizing the classes. This is passed through
        ``easy_params`` to the ``BandStructure`` and ``Conductivity``.
    bounds : Mapping
        Bounds for the fitting parameters. This mapping has the same
        structure as ``init_params``, but only containing the variables
        that are to be fitted, and their values in the mapping must be
        a collection of the form (min, max).
    x_shift : Mapping, optional
        If provided, the y values will be shifted by the y value at
        this x point. The mapping must have the same structure as
        ``x_data``, but with single values instead of arrays as
        the values.
    x_normalize : Mapping, optional
        If provided, the y values will be normalized by the y value
        at this x point. Note that shifting will always be done
        before normalization. The mapping must have the same structure
        as ``x_data``, but with single values instead of arrays as
        the values. Note that shifts are applied before normalization.
    save_path : str, optional
        The directory where the fitting results will be saved.
        If not provided, results will not be saved.
    save_label : str, optional
        Label of the results. If not provided, will be set to
        ``f"y_label_x_label"``. If ``y_label` or ``x_label`` are
        collections of string, they will be joined with an underscore.
    worker_percentage : float, optional
        The percentage of available workers to use for parallel
        computation. If set to 0, it will not be used for setting the
        number of workers. The number of workers can also be set by
        the `workers` keyword argument of differential evolution.
    log_format : str, optional
        The format for logging parameter values.
    **kwargs : dict, optional
        Additional keyword arguments passed to
        `scipy.optimize.differential_evolution`.
    """
    if save_label is None:
        x_string = x_label if isinstance(x_label, str) else "_".join(x_label)
        y_string = y_label if isinstance(y_label, str) else "_".join(y_label)
        save_label = f"{y_string}_{x_string}"
    if worker_percentage > 0:
        if "workers" in kwargs:
            raise ValueError(
                "Cannot set both `worker_percentage` and `workers`.")
        kwargs["workers"] = int(np.ceil(worker_percentage / 100 * cpu_count()))

    update_keys = _extract_flat_keys(bounds, bounds=True)
    bounds = [_extract_flat_value(bounds, key) for key in update_keys]
    x0 = [_extract_flat_value(init_params, key) for key in update_keys]
    for i in range(len(update_keys)):
        if x0[i] is None:
            x_min, x_max = bounds[i]
            x0[i] = (x_min + x_max) / 2

    begin_time = datetime.now()
    cond = _prebuild_cond(init_params)
    fitter = FittingRoutine(init_params, save_path, save_label,
                            update_keys=update_keys)
    result = differential_evolution(
        fitter.residual, bounds=bounds, x0=x0, callback=fitter.log,
        args=(update_keys, x_data, y_data, x_shift, x_normalize, cond),
        **kwargs)
    end_time = datetime.now()

    print(result.message)
    return _save_fit_result(
        result, init_params, update_keys, begin_time, end_time,
        save_path, save_label)


def _prebuild_cond(init_params):
    band = BandStructure(**easy_params(init_params))
    band.discretize()
    cond = Conductivity(band, **easy_params(init_params))
    cond._build_elements()
    cond._build_differential_operator()
    return cond


def _extract_flat_keys(params, bounds=False):
    """Extract dots-separated keys from a nested structure.
    
    This function recursively extracts keys from a nested dictionary
    containing dictionaries and lists, and returns a flat list of keys.
    Each key is represented as a string with dot-separated keys and
    indices, e.g. ``"band_params.a"``, is ``params["band_params"]["a"]``
    or ``"scattering_params.nu.0"`` is
    ``params["scattering_params"]["nu"][0]``.

    Parameters
    ----------
    params : Mapping
        The nested dictionary from which to extract keys.
    
    Returns
    -------
    list[str]
        A list of flattened keys, where each key is a string
        representing the path to the value in the nested structure.
    """
    keys = []
    if isinstance(params, Mapping):
        for key in params:
            value = params[key]
            if _is_value_nested(value, bounds):
                for val_key in _extract_flat_keys(value, bounds=bounds):
                    keys.append(f"{key}.{val_key}")
            else:
                keys.append(key)
    elif isinstance(params, Sequence):
        for (i, value) in enumerate(params):
            if _is_value_nested(value, bounds):
                for val_key in _extract_flat_keys(value, bounds=bounds):
                    keys.append(f"{i}.{val_key}")
            else:
                keys.append(str(i))
    return keys


def _is_value_nested(value, bounds):
    if bounds:
        if isinstance(value, Mapping):
            return any(isinstance(v, Collection) and
                       not isinstance(v, str) for v in value.values())
        elif isinstance(value, Collection):
            return any(isinstance(v, Collection) and
                       not isinstance(v, str) for v in value)
    else:
        return isinstance(value, Collection) and not isinstance(value, str)


def _extract_flat_value(params: Mapping, flat_key: str):
    value = params
    key_parts = flat_key.split('.')
    while key_parts:
        key = key_parts.pop(0)
        if str.isnumeric(key):
            key = int(key)
            if key >= len(value):
                return None
        elif key not in value:
            return None
        value = value[key]
    return value


def _update_flat_value(params: Mapping, flat_key: str, value):
    level_params = params
    key_parts = flat_key.split('.')
    while len(key_parts) > 1:
        key = key_parts.pop(0)
        if str.isnumeric(key):
            key = int(key)
            if key >= len(level_params):
                return
        elif key not in level_params:
            return
        level_params = level_params[key]
    key = key_parts[0]
    if str.isnumeric(key):
        key = int(key)
        if key >= len(level_params):
            return
    level_params[key] = value


def _build_params_from_flat(param_keys, param_values):
    """Build a nested dictionary from flattened keys and values.

    Parameters
    ----------
    param_keys : Collection[str]
        The "flattened" keys of the parameters. See ``extract_keys``
        for more information.
    param_values : Sequence
        The values corresponding to the keys in ``param_keys``.

    Returns
    -------
    dict
        A nested dictionary where the keys are the flattened keys
        and the values are the corresponding values from ``param_values``.
    """
    params = dict()
    param_values = list(param_values)
    for key in param_keys:
        key_parts = key.split('.')
        key = key_parts[0]
        level_params = params
        for part in key_parts[1:]:
            if str.isnumeric(part):
                part = int(part)
                if isinstance(level_params, dict):
                    if key not in level_params:
                        level_params[key] = []
                if isinstance(level_params, list):
                    if level_params[key] == dict():
                        level_params[key] = []
                while part >= len(level_params[key]):
                    level_params[key].append(dict())
            elif key not in level_params:
                level_params[key] = dict()
            level_params = level_params[key]
            key = part
        level_params[key] = param_values.pop(0)
    return params


def _get_y(cond, x_data, y_data, name, y_label_i, y_label_j):
    y = {}
    for label, x in x_data.items():
        setattr(cond, label, x)
    if 'rho' in name.values():
        cond.calculate()
        rho = np.linalg.inv(cond.sigma)
    else:
        cond.calculate(sorted(set(y_label_i.values())),
                       sorted(set(y_label_j.values())))
    for label in y_data:
        if name[label] == 'sigma':
            y[label] = cond.sigma[y_label_i[label], y_label_j[label]]
        elif name[label] == 'rho':
            y[label] = rho[y_label_i[label], y_label_j[label]]
        else:
            raise ValueError(f"Unknown y_data key: {name[label]}")
    return y


def _save_fit_result(result, init_params, update_keys, begin_time,
                     end_time, save_path, save_label):
    result = _result_to_serializable(result)
    result['fit_params'] = _build_params_from_flat(update_keys, result['x'])
    result['residual'] = result['fun']
    result['evaluations'] = result['nfev']
    result['iterations'] = result['nit']
    result.pop('x')
    result.pop('population')
    result.pop('population_energies')
    result.pop('fun')
    result.pop('evalulations')

    result['init_params'] = _build_params_from_flat(
        update_keys, [_extract_flat_value(init_params, key)
                      for key in update_keys])
    all_keys = _extract_flat_keys(init_params, bounds=False)
    fixed_keys = set(all_keys) - set(update_keys)
    result['fixed_params'] = _build_params_from_flat(
        fixed_keys, [_extract_flat_value(init_params, key)
                     for key in fixed_keys])

    result['begin_time'] = begin_time.isoformat()
    result['end_time'] = end_time.isoformat()
    result['runtime'] = (end_time - begin_time).total_seconds()

    if save_path is not None:
        path = Path(save_path) / f"{save_label}.json"
        with path.open('w') as f:
            json.dump(result, f, indent=2)
    return result


def _result_to_serializable(result):
    serializable = {}
    for key, value in result.items():
        if isinstance(value, np.ndarray):
            serializable[key] = value.tolist()
        elif isinstance(value, np.floating):
            serializable[key] = float(value)
        elif isinstance(value, np.integer):
            serializable[key] = int(value)
        else:
            try:
                json.dumps(value)  # test serializability
                serializable[key] = value
            except (TypeError, OverflowError):
                serializable[key] = str(value)
    return serializable



from .integrate import adaptive_octree_integrate

import numpy as np
import sympy
import scipy.sparse
from skimage.measure import marching_cubes

from typing import Union
from collections.abc import Sequence

from scipy.constants import hbar, eV, angstrom
# conversion from energy gradient units to m/s for velocity
velocity_units = 1e-3 * eV * angstrom / hbar


class BandStructure:
    """Contains bandstructure information for a given material.

    In addition to the dispersion relation and general parameters, this
    class also contains methods for discretizing the Fermi surface and
    calculating electronic properties.

    Parameters
    ----------
    dispersion : str
        The dispersion relation. Expresses the dispersion relation
        in terms of symbols in ``wavevector_names`` and additional
        parameters in ``band_params``. It must be parsable and
        differentiable by ``sympy``. Energy units are milli eV.
    chemical_potential : float
        The chemical potential in milli eV.
    unit_cell : Sequence[float]
        The dimensions of the unit cell in angstrom.
    band_params : dict, optional
        The parameters of the dispersion relation. Energy units are
        milli eV and distance units are angstrom.
    domain_size : Sequence[float]
        The ratio of the reciprocal space domain sidelengths to simple
        cubic unit cell dimensions in reciprocal space. The product
        of the numbers in this collection must be equal to the number
        of atoms in the conventional unit cell specified by
        ``unit_cell``.
    periodic : bool or int or Sequence[bool] or Sequence[int]
        If bool, whether periodic boundary conditions are applied to all
        axes or not. If a single int, specifies which single axis is 
        periodic. If a collection, specifies which axes are periodic.
        If the collection is of integers, the integers specify the
        periodic axes, e.g. ``[0, 2]`` means periodic in `x` and `z`
        axes. If the collection is of booleans, it specifies whether
        each axis is periodic or not, e.g. ``[True, False, True]`` means
        periodic in `x` and `z` axes, but not in `y` axis.
    axis_names : str or Sequence[str], optional
        The names of the unit cell axes. Must be parsable by
        `sympy.symbols`.
    wavevector_names : str or Sequence[str], optional
        The names of the wavevector components. Must be parsable by
        `sympy.symbols`.
    resolution :  int or Sequence[int], optional
        Controls the resolution of the grids used for discretizing the
        Fermi surface. If a collection of integers is provided, each
        element corresponds to the resolution along the respective
        axis. If a single integer is provided, it is used for all axes.
    ncorrect : int, optional
        The number of correction steps for improving the accuracy of
        the discretization of the Fermi surface.
    sort_axis : int, optional
        The axis along which to sort the points after triangulation.
        If None, do not sort the points.

    Attributes
    ----------
    dispersion : str
        The dispersion relation. Updating this will automatically
        update ``energy_func`` and ``velocity_func``.
    chemical_potential : float
        The chemical potential in milli eV.
    unit_cell : Sequence[float]
        The dimensions of the unit cell in angstrom.
    domain_size : Sequence[float]
        The ratio of the reciprocal space domain sidelengths to simple
        cubic unit cell dimensions in reciprocal space. The product
        of the numbers in this collection must be equal to the number
        of atoms in the conventional unit cell specified by
        ``unit_cell``.
    periodic : Sequence[int]
        The periodic axes.
    band_params : dict
        The parameters of the dispersion relation. Energy units are
        milli eV and distance units are angstrom.
    kpoints : (N, 3) numpy.ndarray
        The discretized k-points on the Fermi surface. Each row
        corresponds to a k-point in the form ``[kx, ky, kz]``.
    kfaces : (F, 3) numpy.ndarray
        The faces of the triangulated surface in k-space. Each row
        corresponds to a face in the form ``[i, j, k]``, where
        ``i``, ``j``, and ``k`` are the indices of the vertices of
        the face in the ``kpoints`` array.
    periodic_projector : scipy.sparse.csr_array
        Projects quantities into the periodic k-space, where points that
        are periodic images of each other are mapped to the same point.
    resolution : int or Sequence[int]
        The resolution of the grids used for approximating the Fermi
        surface geometry with the marching cubes algorithm.
    ncorrect : int
        The number of Newton--Raphson steps applied to correct the
        triangulated surface after the marching cubes algorithm.
    sort_axis : int or None
        The axis along which to sort the points after triangulation.
    axis_names : str or Sequence[str]
        The names of the unit cell axes.
    wavevector_names : str or Sequence[str]
        The names of the wavevector components.
    """
    def __init__(
            self, dispersion: str, chemical_potential: float,
            unit_cell: Sequence[float], band_params: dict = {},
            domain_size: Sequence[float] = [1.0, 1.0, 1.0],
            periodic: Union[bool, Sequence[Union[int, bool]]] = True,
            axis_names: Union[Sequence[str], str] = ['a', 'b', 'c'],
            wavevector_names: Union[Sequence[str], str] = ['kx', 'ky', 'kz'],
            resolution: Union[int, Sequence[int]] = 21, n_correct: int = 2,
            sort_axis: int = None, **kwargs):
        # avoid triggering the __setattr__ method for the first time
        super().__setattr__('dispersion', dispersion)
        self.band_params = band_params
        self.chemical_potential = chemical_potential
        self.unit_cell = unit_cell
        self.domain_size = domain_size
        self.periodic = periodic
        self.axis_names = axis_names
        self.wavevector_names = wavevector_names
        self.resolution = resolution
        self.n_correct = n_correct
        self._parse_dispersion()
        self.kpoints = None
        self.kfaces = None
        self.periodic_projector = None
        self.sort_axis = sort_axis

    def __setattr__(self, name, value):
        if name == 'dispersion':
            self._parse_dispersion()
        if name == 'resolution':
            if isinstance(value, Sequence):
                value = np.array(value)
            else:
                value = np.array([value, value, value])
        if name in {'unit_cell', 'domain_size'}:
            value = np.array(value, dtype=float)
        if name == 'periodic':
            if isinstance(value, bool):
                if value:
                    value = [0, 1, 2]
                else:
                    value = []
            if isinstance(value, int):
                value = [value]
            elif isinstance(value, Sequence) and all(
                    isinstance(i, bool) for i in value):
                value = [i for i, v in enumerate(value) if v]
        super().__setattr__(name, value)
    
    def __getstate__(self):
        """Get the state of the object for pickling."""
        state = self.__dict__.copy()
        # remove the full energy and velocity functions
        state['_energy_func_full'] = None
        state['_velocity_funcs_full'] = None
        return state
    
    def __setstate__(self, state):
        """Set the state of the object after unpickling."""
        self.__dict__.update(state)
        # re-parse the dispersion relation to restore the full functions
        self._parse_dispersion()
    
    def discretize(self):
        """Discretize the Fermi surface.

        First, the surface is triangulated using the marching cubes
        algorithm with ``resolution`` controlling the resolution of the
        grid. Next, to improve the accuracy of the isosurface,
        ``n_correct`` steps of the Newton--Raphson root-finding method
        are applied to the output of marching cubes. Finally, after the
        surface construction, periodic boundary conditions are applied
        to "stitch" the open ends of the surface together.
        """
        self._gvec = self.domain_size * np.pi / self.unit_cell
        self.kpoints, self.kfaces, _, _ = marching_cubes(
            self.energy_func(*np.mgrid[
                -self._gvec[0]:self._gvec[0]:1j*self.resolution[0],
                -self._gvec[1]:self._gvec[1]:1j*self.resolution[1],
                -self._gvec[2]:self._gvec[2]:1j*self.resolution[2]]),
            level=self.chemical_potential)
        self.kpoints *= (2*self._gvec / (self.resolution-1))[None, :]
        self.kpoints -= self._gvec[None, :]
        for _ in range(self.n_correct):
            self.kpoints = self._apply_newton_correction(self.kpoints)
        if self.sort_axis:
            self._sort_and_reindex(self.sort_axis)
        self._stitch_periodic_boundaries()

    def calculate_filling_fraction(self, depth: int = 7) -> float:
        """Calculate the filling fraction n of the material.

        The filling fraction is calculated by integrating the volume
        in the reciprocal space having energy bellow the Fermi level
        (calculated by an adaptive octree integration method), then
        dividing by the volume of the unit cell in the reciprocal space.

        Parameters
        ----------
        depth : int, optional
            The depth of the adaptive octree integration. Higher values
            result in more accurate integration, but take exponentially
            longer to compute.

        Returns
        -------
        float
            The filling fraction n of the material.
        """
        self._gvec = self.domain_size * np.pi / self.unit_cell
        # the extra factor of 2 is the spin degeneracy
        return 2 * adaptive_octree_integrate(
            lambda kx, ky, kz: (self.energy_func(kx, ky, kz)
                                < self.chemical_potential),
            (-self._gvec[0], self._gvec[0], -self._gvec[1], self._gvec[1],
             -self._gvec[2], self._gvec[2]), depth=depth
            ) / 8 / np.prod(self._gvec)

    def calculate_electron_density(self, depth: int = 7) -> float:
        """Calculate the electron density n_e of the material.

        Note that the surface needs to be discretized before calling
        this method. First, the filling fraction is calculated (see
        ``calculate_filling_fraction``). The electron density is
        obtained by dividing the filling fraction by the volume of
        the unit cell in real space.

        Parameters
        ----------
        depth : int, optional
            The depth of the adaptive octree integration in
            ``calculate_filling_fraction``.

        Returns
        -------
        float
            The electron density n_e of the material in SI units.
        """
        filling_fraction = self.calculate_filling_fraction(depth)
        # the volume of the unit cell in real space is scaled by
        # the inverse scaling of the unit cell in reciprocal space
        unit_cell_volume = (np.prod(self.unit_cell) * angstrom**3
                            / np.prod(self.domain_size))
        return filling_fraction / unit_cell_volume

    def calculate_mass(self):
        """Calculate the effective mass of the charge carries.

        Returns
        -------
        float
            The effective mass divided by the rest mass of
            the electron, m_e.
        """
        # Placeholder for actual calculation
        return 0.0
    
    def energy_func(self, kx, ky, kz):
        """Calculate the energy at the given k-point.

        Parameters
        ----------
        kx, ky, kz : float
            The components of the wavevector in angstrom^-1.

        Returns
        -------
        object like kx, ky, kz
            The energy at the given k-point in milli eV.
        """
        return self._energy_func_full(
            kx, ky, kz, *self.unit_cell, **self.band_params)
    
    def velocity_func(self, kx, ky, kz):
        """Calculate the velocity at the given k-point.

        Parameters
        ----------
        kx, ky, kz : float
            The components of the wavevector in angstrom^-1.

        Returns
        -------
        list of 3 objects like kx, ky, kz
            The velocity vector at the given k-point in m/s.
        """
        return [vfunc(kx, ky, kz, *self.unit_cell, **self.band_params)
                for vfunc in self._velocity_funcs_full]

    def _parse_dispersion(self):
        """
        Parse the dispersion relation and extract the necessary
        information for further calculations.
        """
        ksymbols = sympy.symbols(self.wavevector_names)
        all_symbols = (ksymbols + sympy.symbols(self.axis_names)
                       + sympy.symbols(list(self.band_params.keys())))
        # symbolic expressions
        self._energy_sympy = sympy.sympify(self.dispersion)
        self._velocities_sympy = [
            sympy.diff(self._energy_sympy, k) * velocity_units
            for k in sympy.symbols(self.wavevector_names)]
        # replace zero velocities with zero arrays
        for i, v in enumerate(self._velocities_sympy):
            if v == 0:
                self._velocities_sympy[i] = f"numpy.zeros_like({ksymbols[i]})"
        # convert expressions into python functions
        self._energy_func_full = sympy.lambdify(
            all_symbols, self._energy_sympy)
        self._velocity_funcs_full = [
            sympy.lambdify(all_symbols, vexpr, 'numpy')
            for vexpr in self._velocities_sympy]

    def _sort_and_reindex(self, sort_axis):
        new_order, self.kfaces = self._generate_reindex(sort_axis)
        self.kpoints = self.kpoints[new_order]
    
    def _generate_reindex(self, sort_axis):
        new_order = np.argsort(self.kpoints[:, sort_axis])
        old_to_new_map = np.empty(len(new_order), dtype=int)
        old_to_new_map[new_order] = np.arange(len(new_order))
        return new_order, old_to_new_map[self.kfaces]

    def _stitch_periodic_boundaries(self):
        """
        Find duplicate points on the periodic boundaries, then make the
        periodic mesh arrays.
        """
        duplicates = dict()
        threshold = np.min(self._gvec / self.resolution) / 10
        for axis in self.periodic:
            low_border = np.argwhere(
                self.kpoints[:, axis] + self._gvec[axis] < threshold).ravel()
            high_border = np.argwhere(
                self.kpoints[:, axis] - self._gvec[axis] > -threshold).ravel()
            if len(low_border) == 0 or len(high_border) == 0:
                continue

            min_dist = min(self._get_min_border_distance(low_border),
                           self._get_min_border_distance(high_border))

            k1 = self.kpoints[low_border][None, :]
            k2 = self.kpoints[high_border][:, None]
            kdiff = k2 - k1
            kdiff[:, :, axis] += self._gvec[axis]
            kdiff[:, :, axis] %= 2 * self._gvec[axis]
            kdiff[:, :, axis] -= self._gvec[axis]
            kdiff = np.linalg.norm(kdiff, axis=-1)

            min_pair = np.argmin(kdiff, axis=1)
            is_duplicate = kdiff[np.arange(len(high_border)), min_pair
                                 ] < min_dist / 2
            duplicates.update(dict(zip(
                high_border[is_duplicate],
                low_border[min_pair[is_duplicate]])))
        self._build_periodic_projector(duplicates)
    
    def _get_min_border_distance(self, border):
        """
        Find minimum intra-layer distance to set the threshold
        for duplicate point detection.
        """
        is_triangle_point_in_border = np.isin(self.kfaces, border)
        border_triangles = self.kfaces[np.any(
            is_triangle_point_in_border, axis=1)]
        points = self.kpoints[border_triangles]
        is_triangle_point_in_border = is_triangle_point_in_border[
            np.any(is_triangle_point_in_border, axis=1)]
        is_pair_intra_layer = np.logical_xor(
            is_triangle_point_in_border,
            np.roll(is_triangle_point_in_border, 1, axis=-1))
        return np.min(np.linalg.norm(
            (points - np.roll(points, 1, axis=1))[is_pair_intra_layer],
            axis=-1))

    def _build_periodic_projector(self, duplicates):
        """
        Build the periodic kpoints and kfaces arrays by removing
        duplicate points and reindexing.
        """
        if not duplicates:
            self.periodic_projector = scipy.sparse.eye(
                len(self.kpoints), format='csr')
        else:
            unique_mask = np.full(len(self.kpoints), True)
            unique_mask[list(duplicates.keys())] = False
            reindex_map = np.cumsum(unique_mask) - 1
            reindex_map[list(duplicates.keys())] = reindex_map[
                list(duplicates.values())]
            self.periodic_projector = scipy.sparse.csr_array(
                (np.ones(len(self.kpoints)),
                (reindex_map, np.arange(len(self.kpoints)))),
                shape=(np.count_nonzero(unique_mask), len(self.kpoints)))

    def _apply_newton_correction(self, points):
        residuals = self.energy_func(
            points[:, 0], points[:, 1], points[:, 2]) - self.chemical_potential
        gradients = np.column_stack(self.velocity_func(
            points[:, 0], points[:, 1], points[:, 2])) / velocity_units
        gradient_norms = np.linalg.norm(gradients, axis=-1)
        return points - (residuals/gradient_norms**2)[:, None]*gradients



import unittest
import elecboltz
import tempfile
import pathlib
import numpy as np


class TestLoader(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.addCleanup(self.temp_dir.cleanup)
        with open(pathlib.Path(self.temp_dir.name)
                  / 'test_data_phi30_B_10.0_rho0=0.5_date_11072025.csv',
                  'w') as f:
            f.write("# these\n"
                    "# are\n"
                    "# comments\n"
                    "theta,rho_a,rho_c\n"
                    "0.5,0.01,0.1\n"
                    "1.0,0.02,0.2\n"
                    "1.5,0.03,0.3")
        with open(pathlib.Path(self.temp_dir.name)
                  / 'test_data_phi=45_B2.4_rho0=1.0_date12072025.csv',
                  'w') as f:
            f.write("# more\n"
                    "# comments\n"
                    "# here\n"
                    "theta,rho_a,rho_c\n"
                    "2.0,0.04,0.4\n"
                    "2.5,0.05,0.5\n"
                    "3.0,0.06,0.6\n"
                    "3.5,0.07,0.7\n"
                    "4.0,0.08,0.8\n")
        with open(pathlib.Path(self.temp_dir.name)
                  / 'test_data_phi=15_B3.0_rho0=1.5_date=13072025.csv',
                  'w') as f:
            f.write("# even\n"
                    "# more\n"
                    "# comments\n"
                    "theta,rho_a,rho_c\n"
                    "4.5,0.09,0.9\n"
                    "5.0,0.1,1.0\n")
        with open(pathlib.Path(self.temp_dir.name)
                  / 'not_the_data_phi=15_B3.0_rho0=1.5.csv', 'w') as f:
            f.write("# yet\n"
                    "# more\n"
                    "# comments\n"
                    "theta,rho_a,rho_c\n"
                    "5.5,0.11,1.1\n"
                    "6.0,0.12,1.2\n")

    def test_auto_search(self):
        loader = elecboltz.Loader(
            save_new_labels=True, save_new_values=True, data_type='plain')
        loader.load(folder_path=str(self.temp_dir.name),
                    prefix='test_data_', delimiter=',', skiprows=4)
        self.assertTrue(all(
            np.all(x == x_test) for x, x_test in zip(loader.x_data_raw['theta'],
                [np.array([0.5, 1.0, 1.5]),
                 np.array([4.5, 5.0]),
                 np.array([2.0, 2.5, 3.0, 3.5, 4.0])])),
            "x_data_raw does not match expected values.")
        self.assertTrue(all(
            np.all(y == y_test) for y, y_test in
            zip(loader.y_data_raw['rho_a'],
                [np.array([0.01, 0.02, 0.03]),
                 np.array([0.09, 0.1]),
                 np.array([0.04, 0.05, 0.06, 0.07, 0.08])])),
            "y_data_raw does not match expected values.")
        self.assertTrue(all(
            np.all(y == y_test) for y, y_test in
            zip(loader.y_data_raw['rho_c'], [np.array([0.1, 0.2, 0.3]),
                                       np.array([0.9, 1.0]),
                                       np.array([0.4, 0.5, 0.6, 0.7, 0.8])])),
            "y_data_raw does not match expected values.")
        self.assertTrue(all(
            np.all(x == x_test) for x, x_test in zip(
                [loader.x_data['theta'], loader.x_data['phi'],
                 loader.x_data['B'], loader.x_data['rho0']],
                [np.array([0.5, 1.0, 1.5, 4.5, 5.0, 2.0, 2.5, 3.0, 3.5, 4.0]),
                 np.array([30.0, 30.0, 30.0, 15.0, 15.0,
                           45.0, 45.0, 45.0, 45.0, 45.0]),
                 np.array([10.0, 10.0, 10.0, 3.0, 3.0,
                           2.4, 2.4, 2.4, 2.4, 2.4])])),
            "x_data does not match expected values.")
        self.assertTrue(all(
            np.all(y == y_test) for y, y_test in zip(
                [loader.y_data['rho_a'], loader.y_data['rho_c']],
                [np.array([0.01, 0.02, 0.03, 0.09, 0.1,
                           0.04, 0.05, 0.06, 0.07, 0.08]),
                 np.array([0.1, 0.2, 0.3, 0.9, 1.0,
                           0.4, 0.5, 0.6, 0.7, 0.8])])),
            "y_data does not match expected values.")
        pass
    
    def test_plain_load(self):
        loader = elecboltz.Loader(
            x_vary_label='theta',
            x_search={'phi': [30, 45], 'B': [10.0, 2.4], 'rho0': [0.5, 1.0]},
            y_label=['rho_xx', 'rho_xy'], data_type='plain')
        loader.load(folder_path=str(self.temp_dir.name), prefix='test_data_',
                    x_columns=[0], y_columns=[1, 2], delimiter=',', skiprows=4)
        self.assertTrue(all(np.all(x == x_test)
            for x, x_test in zip(loader.x_data_raw['theta'],
                [np.array([0.5, 1.0, 1.5]),
                 np.array([2.0, 2.5, 3.0, 3.5, 4.0])])),
            "x_data_raw does not match expected values.")
        self.assertTrue(all(
            np.all(y == y_test) for y, y_test in
            zip(loader.y_data_raw['rho_xx'],
                [np.array([0.01, 0.02, 0.03]),
                 np.array([0.04, 0.05, 0.06, 0.07, 0.08])])),
            "y_data_raw does not match expected values.")
        self.assertTrue(all(
            np.all(y == y_test) for y, y_test in
            zip(loader.y_data_raw['rho_xy'], [np.array([0.1, 0.2, 0.3]),
                                       np.array([0.4, 0.5, 0.6, 0.7, 0.8])])),
            "y_data_raw does not match expected values.")
        self.assertTrue(all(
            np.all(x == x_test) for x, x_test in zip(
                [loader.x_data['theta'], loader.x_data['phi'],
                 loader.x_data['B'], loader.x_data['rho0']],
                [np.array([0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0]),
                 np.array([30.0, 30.0, 30.0, 45.0, 45.0, 45.0, 45.0, 45.0]),
                 np.array([10.0, 10.0, 10.0, 2.4, 2.4, 2.4, 2.4, 2.4]),
                 np.array([0.5, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0])])),
            "x_data does not match expected values.")
        self.assertTrue(all(
            np.all(y == y_test) for y, y_test in zip(
                [loader.y_data['rho_xx'], loader.y_data['rho_xy']],
                [np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08]),
                 np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])])),
            "y_data does not match expected values.")

    def test_admr_load(self):
        loader = elecboltz.Loader(
            x_vary_label='theta',
            x_search={'phi': [30, 45], 'B': [10.0, 2.4], 'rho0': [0.5, 1.0]},
            y_label=['rho_xx', 'rho_xy'], data_type='admr')
        loader.load(folder_path=str(self.temp_dir.name), prefix='test_data_',
                    x_columns=[0], y_columns=[1, 2], delimiter=',', skiprows=4)
        theta = np.deg2rad(np.array([0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0]))
        phi = np.deg2rad(
            np.array([30.0, 30.0, 30.0, 45.0, 45.0, 45.0, 45.0, 45.0]))
        B = np.array([10.0, 10.0, 10.0, 2.4, 2.4, 2.4, 2.4, 2.4])
        self.assertEqual(list(loader.x_data.keys()), ['field'])
        self.assertTrue(np.all(loader.x_data['field']
            == B[:, None] * np.column_stack((np.sin(theta) * np.cos(phi),
                                             np.sin(theta) * np.sin(phi),
                                             np.cos(theta)))),
            "x data does not match expected values.")
        pass


if __name__ == '__main__':
    unittest.main()



import numpy as np
import re
from typing import Mapping, Sequence, Union
from pathlib import Path
from collections import defaultdict


class Loader:
    """Convenience class for loading data split into multiple files.
    
    Prepares the data for ``fit_model`` by processing multiple files
    labeled with values of independent variables and containing the
    variation of one variable. For example, the files can be labeled
    with the values of ``phi`` and ``field``, and contain the variation
    of ``theta``.

    Parameters
    ----------
    x_vary_label : Union[str, Sequence[str]], optional
        Label of the independent variable(s) that varies inside the
        files. If not provided, it will be inferred from the file
        contents (column headers).
    x_search : Mapping[str, Sequence[Union[int, float]]], optional
        Dictionary mapping labels of independent variables to the values
        to be searched for in the file names. Example: 
        ``{'phi': [30, 60], 'field': [1.3, 2.7]}``.
    y_label : Sequence[str], optional
        Labels of the dependent variables, used for fitting routine.
        See ``elecboltz.fit.fit_model`` for details about the allowed
        labels. If None, will be inferred from the file contents
        (column headers).
    save_new_labels : bool, optional
        If True, new labels found in the file names will be saved to
        ``x_search``.
    save_new_values : bool, optional
        If True, new values for existing labels in ``x_search`` will be
        added to the list of values for that label.
    data_type : {'plain', 'admr'}, optional
        Type of data to load. If ``'plain'``, no processing is done on
        the data and the x_data is constructed directly from the
        collected values. If ``'admr'``, the x_data is constructed
        such that it contains the field vector at each index.
    
    Attributes
    ----------
    x_data : defaultdict[str, list[np.ndarray]]
        x_data variable for the fitting procedure.
    y_data : defaultdict[str, list[np.ndarray]]
        y_data variable for the fitting procedure.
    x_data_raw : defaultdict[str, list[np.ndarray]]
        Raw data of the independent variable(s) collected from the
        files. Each ``x_vary_label`` is mapped to the corresponding
        arrays. Each array corresponds to a different value of the
        labels in ``x_search``.
    y_data_raw : deefaultdict[str, list[np.ndarray]]
        Raw data of the dependent variable collected from the files.
        Each sequence corresponds to a different dependent variable
        in ``y_label``, and each array inside that corresponds
        to a different value in ``x_search_values``.
    x_data_interpolated : defaultdict[str, list[np.ndarray]]
        Interpolated, but unprocessed, data of the independent variable
        varying inside the files.
    y_data_interpolated : defaultdict[str, list[np.ndarray]]
        Interpolated, but unprocessed, data of the dependent variable
        collected from the files.
    x_vary_label : Union[str, Sequence[str]], optional
        Label of the independent variable(s) that varies inside the
        files.
    x_search: Mapping[str, Union[int, float]]
        Dictionary mapping labels of independent variables inside file
        names to their values.
    """
    def __init__(self, x_vary_label: Union[str, Sequence[str]] = None,
                 x_search: Mapping[str, Sequence[Union[int, float]]] = {},
                 y_label: Sequence[str] = None, save_new_labels: bool = False,
                 save_new_values: bool = False, data_type: str = 'admr'):
        self.x_vary_label = x_vary_label
        self.x_search = x_search
        self.y_label = y_label
        self.save_new_labels = save_new_labels
        self.save_new_values = save_new_values
        self.data_type = data_type
        self.x_data = defaultdict(list)
        self.y_data = defaultdict(list)
        self.x_data_raw = defaultdict(list)
        self.y_data_raw = defaultdict(list)
        self.x_data_interpolated = defaultdict(list)
        self.y_data_interpolated = defaultdict(list)

    def __setattr__(self, name, value):
        if name == 'x_search':
            if value != {}:
                values = list(value.values())
                value_len = len(values[0])
                for values in values:
                    if value_len != len(values):
                        raise ValueError("All labels in x_search must have"
                                         " the same number of values.")
        super().__setattr__(name, value)

    def load(self, folder_path: str = '.', prefix: str = '',
             recursive: bool = True,
             x_columns: Union[Sequence[int], Sequence[str]] = None,
             y_columns: Union[Sequence[int], Sequence[str]] = None,
             x_units: Union[Sequence[float], float] = 1.0,
             y_units: Union[Sequence[float], float] = 1.0,
             **kwargs):
        """Load the data from files in the specified folder.

        The function automatically determines which files to load based
        on the given parameters. The file names should be underline
        separated, with each part representing a different
        independent variable. You can put an equal sign or an underline
        between the variable name and its value, but it is not necessary
        if there is no ambiguity. If there is ambiguity, you must put an
        equal sign (an underline is not sufficient). For example, the
        file name can be `admr_lsco_phi=30_field10_rho0=0.1.csv` or
        `admr_lsco_phi_30_field=10_rho0=0.1.csv`.

        Parameters
        ----------
        folder_path : str, optional
            Path to the folder containing the data files.
        prefix : str, optional
            Prefix for the data files.
        recursive : bool, optional
            If True, search for files recursively in the folder and its
            subfolders. If False, search only in the specified folder.
        x_columns : Union[Sequence[int], Sequence[str]], optional
            Override which columns to load for the independent variable.
            If a sequence of integers, it specifies the column indices
            to load. If a sequence of strings, it specifies the column
            names (headers) to load. If not provided, the first column
            is loaded as the independent variable.
        y_columns : Union[Sequence[int], Sequence[str]], optional
            Override which columns to load for the dependent variables.
            If a sequence of integers, it specifies the column indices
            to load. If a sequence of strings, it specifies the column
            names (headers) to load. If not provided, all columns except
            the first one (independent variable) are loaded.
        x_units : Union[Sequence[float], float], optional
            Units for the independent variable(s). If a single float is
            provided, it is applied to all independent variables. If a
            sequence is provided, it must match the number of independent
            variables.
        y_units : Union[Sequence[float], float], optional
            Units for the dependent variable(s). If a single float is
            provided, it is applied to all dependent variables. If a
            sequence is provided, it must match the number of dependent
            variables.
        **kwargs : dict, optional
            Additional keyword arguments to pass to ``numpy.loadtxt``.
        """
        if recursive:
            files = sorted(Path(folder_path).rglob(f"{prefix}*"))
        else:
            files = sorted(Path(folder_path).glob(f"{prefix}*"))
        for file in files:
            if file.is_dir() or not file.name.startswith(prefix):
                continue

            label_map = _extract_labels_and_values(file.name)
            for label, value in label_map.items():
                if self.save_new_labels and label not in self.x_search:
                    self.x_search[label] = []
                if self.save_new_values:
                    self.x_search[label].append(value)

            if self.x_search != {}:
                if not all(label in label_map for label in self.x_search):
                    continue
                if not all(label_map[label] in self.x_search[label]
                           for label in self.x_search):
                    continue
                possible_idx = set()
                for label in self.x_search:
                    if label in label_map:
                        new_idx = [
                            i for i, value in enumerate(self.x_search[label])
                            if value == label_map[label]]
                        if possible_idx == set():
                            possible_idx = set(new_idx)
                        else:
                            possible_idx.intersection_update(new_idx)
                if not possible_idx:
                    continue
                else:
                    idx = min(possible_idx)
            else:
                if self.x_data_raw == {}:
                    idx = 0
                else:
                    first_label = list(self.x_data_raw.keys())[0]
                    idx = len(self.x_data_raw[first_label])

            self._extract_data(file, idx, x_columns, y_columns,
                               x_units, y_units, **kwargs)
        self.process_data()

    def interpolate(self, n_points: int = 50, x_min: float = None,
                    x_max: float = None, x_shift: float = None,
                    x_normalize: float = None, y_shift: float = 0.0,
                    y_normalize: float = 1.0):
        """
        Interpolate the loaded data to the specified number of points.

        For now, only supports linear interpolation for a single
        independent variable varying inside the files.

        Parameters
        ----------
        npoints : int, optional
            Number of points to interpolate to.
        x_min : float, optional
            Lower boundary of the range for the independent variable
            (varying inside the files). If not provided, it is set to
            the minimum value of the independent variable in the loaded
            data.
        x_max : float, optional
            Upper boundary of the range for the independent variable
            (varying inside the files). If not provided, it is set to
            the maximum value of the independent variable in the loaded
            data.
        x_shift : float, optional
            If provided, the data will be shifted by the value at this
            point.
        x_normalize : float, optional
            If provided, the data will be normalized by the value at
            this point. Note that shifts are applied before
            normalization.
        y_shift : float, optional
            The data will be shifted to this value (if ``x_shift``
            is provided).
        y_normalize : float, optional
            The data will be normalized to this value
            (if ``x_normalize`` is provided).
        """
        self.x_data_interpolated = defaultdict(list)
        self.y_data_interpolated = defaultdict(list)
        for i, x in enumerate(self.x_data_raw[self.x_vary_label]):
            x_min = min(x) if x_min is None else x_min
            x_max = max(x) if x_max is None else x_max
            x_new = np.linspace(x_min, x_max, n_points)
            self.x_data_interpolated[self.x_vary_label].append(x_new)
            for y_label, y in self.y_data_raw.items():
                self.y_data_interpolated[y_label].append(
                    np.interp(x_new, x, y[i]))
                if x_shift is not None:
                    y = np.interp(x_shift, x, y[i])
                    self.y_data_interpolated[y_label][-1] -= y
                    self.y_data_interpolated[y_label][-1] += y_shift
                if x_normalize is not None:
                    y = np.interp(x_normalize, x, y[i])
                    self.y_data_interpolated[y_label][-1] /= y
                    self.y_data_interpolated[y_label][-1] *= y_normalize

        self.process_data()

    def process_data(self):
        """Process the loaded data to prepare it for fitting.
        
        Fills the `x_data` and `y_data` attributes with the correct
        values. This is run at the end of the `load` and `interpolate`
        methods. You can use it if you do extra processing on the data
        after loading or interpolating.
        """
        self.x_data = defaultdict(list)
        self.y_data = defaultdict(list)
        if self.y_data_interpolated != {}:
            y_separate = self.y_data_interpolated
        else:
            y_separate = self.y_data_raw
        for label, data in y_separate.items():
            self.y_data[label] = np.concatenate(data)

        if isinstance(self.x_vary_label, str):
            x_vary_label = [self.x_vary_label]
        else:
            x_vary_label = self.x_vary_label
        all_labels = x_vary_label + list(self.x_search.keys())
        x_data_stitched = {label: [] for label in all_labels}
        if self.x_data_interpolated != {}:
            x_separate = self.x_data_interpolated
        else:
            x_separate = self.x_data_raw
        for label in self.x_search:
            for i, value in enumerate(self.x_search[label]):
                x_data_stitched[label].append(
                    np.full(len(x_separate[x_vary_label[0]][i]), value))
        for label in x_vary_label:
            for i, data in enumerate(x_separate[label]):
                while len(x_data_stitched[label]) <= i:
                    x_data_stitched[label].append(np.array([]))
                x_data_stitched[label][i] = data
        x_data_stitched = {label: np.concatenate(data)
                           for label, data in x_data_stitched.items()}
        if self.data_type == 'plain':
            self.x_data = x_data_stitched
        elif self.data_type == 'admr':
            indexing_labels = dict()
            for label in x_data_stitched:
                if label in ['B', 'Bmag', 'Bamp', 'H', 'Hmag', 'Hamp']:
                    indexing_labels['field'] = label
                else:
                    indexing_labels[label.lower()] = label
            field = x_data_stitched[indexing_labels['field']]
            phi = np.deg2rad(x_data_stitched[indexing_labels['phi']])
            theta = np.deg2rad(x_data_stitched[indexing_labels['theta']])
            self.x_data = {'field': field[:, None] * np.column_stack((
                np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi),
                np.cos(theta)))}

    def _extract_data(self, file, idx, x_columns, y_columns,
                      x_units, y_units, **kwargs):
        x_columns, y_columns = self._extract_xy_labels(
            file, x_columns, y_columns, **kwargs)
        data = np.loadtxt(file, **kwargs)

        # "pack" single labels into a list
        if isinstance(self.x_vary_label, str):
            x_vary_label = [self.x_vary_label]
        else:
            x_vary_label = self.x_vary_label
        if isinstance(self.y_label, str):
            y_label = [self.y_label]
        else:
            y_label = self.y_label
        if isinstance(x_units, (int, float)):
            x_units = [x_units] * len(x_columns)
        if isinstance(y_units, (int, float)):
            y_units = [y_units] * len(y_columns)

        for label, col, unit in zip(x_vary_label, x_columns, x_units):
            while len(self.x_data_raw[label]) <= idx:
                self.x_data_raw[label].append(np.array([]))
            self.x_data_raw[label][idx] = unit * data[:, col]
        for label, col, unit in zip(y_label, y_columns, y_units):
            while len(self.y_data_raw[label]) <= idx:
                self.y_data_raw[label].append(np.array([]))
            self.y_data_raw[label][idx] = unit * data[:, col]

    def _extract_xy_labels(self, file, x_columns, y_columns, **kwargs):
        with open(file, 'r') as f:
            for _ in range(kwargs.get('skiprows', 1)):
                line = f.readline()
            while line.startswith(kwargs.get('comments', '#')):
                line = f.readline()
            if self.x_vary_label is None:
                self.x_vary_label = line.split(',')[0].strip()

            headers = [header.strip() for header in line.split(',')]
            if x_columns is None:
                if self.x_vary_label is None:
                    x_columns = [0]
                    self.x_vary_label = headers[0]
                elif isinstance(self.x_vary_label, str):
                    x_columns = [headers.index(self.x_vary_label)]
                else:
                    x_columns = [headers.index(label)
                                 for label in self.x_vary_label]
            elif self.x_vary_label is None:
                self.x_vary_label = [headers[col] for col in x_columns]

            if y_columns is None:
                if self.y_label is None:
                    y_columns = sorted(list(
                        set(range(len(headers))) - set(x_columns)))
                    self.y_label = [headers[col] for col in y_columns]
                elif isinstance(self.y_label, str):
                    y_columns = [headers.index(self.y_label)]
                else:
                    y_columns = [headers.index(label)
                                 for label in self.y_label]
            elif self.y_label is None:
                self.y_label = [headers[col] for col in y_columns]
        return x_columns, y_columns

def _extract_labels_and_values(file_name):
    label_map = {}
    parts = file_name.split('_')
    parts[-1] = '.'.join(parts[-1].split('.')[:-1])
    for i, part in enumerate(parts):
        if '=' in part:
            label, value = part.split('=')
            value = re.search(r'([-+]?\d+\.?\d*)', value).group(0)
            value = float(value)
            if int(value) == value:
                value = int(value)
            label_map[label] = value
        else:
            value = re.search(r'([-+]?\d+\.?\d*[a-zA-Z]*)', part)
            if value:
                value = float(re.search(r'([-+]?\d+\.?\d*)',
                                        value.group(0)).group(0))
                if int(value) == value:
                    value = int(value)
                label = re.search(r'([a-zA-Z_]+)', part)
                if label:
                    label_map[label.group(0)] = value
                elif i > 0:
                    if re.search(r'([a-zA-Z_]+)', parts[i - 1]):
                        label_map[parts[i - 1]] = value
    return label_map



import numpy as np
from typing import Callable, Sequence


def adaptive_octree_integrate(
        func: Callable, bounds: Sequence[float], depth: int = 7) -> float:
    """
    Integrate a binary-valued function over a cubic volume using an
    adaptive octree subdivision scheme.

    Parameters
    ----------
    func : Callable
        A function that takes three arguments ``(x, y, z)`` and returns
        a boolean value indicating whether the point is inside the
        volume of interest.
    bounds : Sequence[float]
        A collection of six floats
        ``(xmin, xmax, ymin, ymax, zmin, zmax)``
        defining the bounds of the cubic volume to integrate over.
    depth : int, optional
        The maximum depth of the octree.

    Returns
    -------
    float
        The estimated integral of the function over the specified volume.
    """
    bounds = np.array([bounds]).transpose()
    integral = 0.0
    for _ in range(depth):
        corners, centers = _get_corners_and_centers(bounds)
        points = np.concatenate([corners, centers], axis=1)
        is_point_inside = func(points[0], points[1], points[2]
                               ).reshape((9, -1))

        is_voxel_inside = np.all(is_point_inside, axis=0)
        inside_bounds = bounds[:, is_voxel_inside]
        added_volume = np.sum(
            np.prod(inside_bounds[1::2] - inside_bounds[::2], axis=0))
        integral += added_volume
        
        is_voxel_partially_inside = np.logical_and(
            np.any(is_point_inside, axis=0), np.logical_not(is_voxel_inside))
        bounds = bounds[:, is_voxel_partially_inside]
        centers = centers[:, is_voxel_partially_inside]
        bounds = _subdivide_voxel(bounds, centers)
    integral += _integrate_final_octree(func, bounds)
    return integral


def _get_corners_and_centers(bounds):
    corners = np.concatenate(
        [np.vstack([bounds[i], bounds[j], bounds[k]])
         for i in range(2) for j in range(2, 4) for k in range(4, 6)],
        axis=1)
    centers = (bounds[::2] + bounds[1::2]) / 2
    return corners, centers

def _subdivide_voxel(bounds, centers):
    return np.concatenate([
        np.vstack([bounds[0], centers[0], bounds[2], centers[1],
                   bounds[4], centers[2]]),
        np.vstack([centers[0], bounds[1], bounds[2], centers[1],
                   bounds[4], centers[2]]),
        np.vstack([bounds[0], centers[0], centers[1], bounds[3],
                   bounds[4], centers[2]]),
        np.vstack([centers[0], bounds[1], centers[1], bounds[3],
                   bounds[4], centers[2]]),
        np.vstack([bounds[0], centers[0], bounds[2], centers[1],
                   centers[2], bounds[5]]),
        np.vstack([centers[0], bounds[1], bounds[2], centers[1],
                   centers[2], bounds[5]]),
        np.vstack([bounds[0], centers[0], centers[1], bounds[3],
                   centers[2], bounds[5]]),
        np.vstack([centers[0], bounds[1], centers[1], bounds[3],
                   centers[2], bounds[5]])],
        axis=1)


def _integrate_final_octree(func, bounds):
    corners, centers = _get_corners_and_centers(bounds)
    volumes = np.prod(bounds[1::2] - bounds[::2], axis=0)
    return np.sum((np.sum(func(corners[0], corners[1], corners[2]
                               ).reshape((8, -1)), axis=0)
                   + func(centers[0], centers[1], centers[2])) / 9
                  * volumes)



from .scattering import build_scattering_function
from copy import deepcopy


def easy_params(params):
    """Convenience function to set parameters for the simulation.

    List of convenience features:

    * | Set unit cell dimensions with named parameters `a`, `b`, and
      | `c`. If `b` or `c` are not given, they are assumed to be equal
      | to `a`.
    * Define an `energy_scale` which scales all energy parameters.
    * | Set the chemical potential with `mu` in `band_params`. If `mu`
      | is set in `band_params`, it is assumed the energy dispersion is
      | shifted by the chemical potential, so the associated variable
      | is set to 0.0 in the returned parameters.
    * | Set the dispersion relation with a default tight-binding model.
      | See `get_tight_binding_dispersion` for the list of parameters
      | and the resulting expression.
    * | Build the scattering function using predefined
      | `scattering_models` and the `scattering_params` associated with
      | them. See `build_scattering_function` for supported scattering
      | models and their parameters. `scattering_models` is assumed to
      | be only one `isotropic` model if not specified.

    Parameters
    ----------
    params : dict
        Simplified (easy-to-use) parameters for the simulation.
    
    Returns
    -------
    dict
        Parameters compatible with the classes in the package.
    """
    new_params = deepcopy(params)
    # unit cell dimensions indicated by axis names
    if 'a' in params:
        unit_cell = [params['a'], params['a'], params['a']]
        if 'b' in params:
            unit_cell[1] = params['b']
        if 'c' in params:
            unit_cell[2] = params['c']
        new_params['unit_cell'] = unit_cell
    # some like to shift the energy dispersion itself by the chemical
    # potential and treat similarly to the other energy parameters
    if 'band_params' in params:
        new_params['band_params'] = params['band_params'].copy()
        if 'mu' in params['band_params']:
            new_params['chemical_potential'] = 0.0
    # scale all energy parameters by a given factor
    if 'energy_scale' in params:
        for key in new_params['band_params']:
            new_params['band_params'][key] *= params['energy_scale']
    # get the default tight-binding dispersion relation
    if 'dispersion' not in new_params:
        new_params['dispersion'] = get_tight_binding_dispersion(
            new_params['band_params'])
    # automatically build the scattering function from named parameters
    if 'scattering_params' in params:
        if 'scattering_models' not in params:
            new_params['scattering_models'] = ['isotropic']
        new_params['scattering_rate'] = build_scattering_function(
            new_params['scattering_params'], new_params['scattering_models'])
    return new_params


def get_tight_binding_dispersion(band_params) -> str:
    """
    Get the tight-binding dispersion relation containing terms relating
    to the parameters in `band_params`.

    The full tight-binding dispersion relation is given by::

        -mu - 2*t * (cos(a*kx)+cos(b*ky))
        - 4*tp * cos(a*kx)*cos(b*ky)
        - 2*tpp * (cos(2*a*kx)+cos(2*b*ky))
        - 2*tz * (cos(a*kx)-cos(b*ky))**2
            * cos(a*kx/2)*cos(b*ky/2)*cos(c*kz/2)

    The list of parameters is as follows:

    * mu: Chemical potential.
    * t: Nearest-neighbor hopping parameter in the x-y plane.
    * tp: Next-nearest-neighbor hopping parameter in the x-y plane.
    * | tpp: Next-next-nearest-neighbor hopping parameter
      | in the x-y plane.
    * | tz: Nearest-neighbor hopping parameter between
      | the different layers in the z direction.

    Parameters
    ----------
    band_params : dict or set
        Dictionary or set of parameters for the tight-binding model.
    
    Returns
    -------
    str
        The dispersion relation expression string
    """
    dispersion = ""
    if 'mu' in band_params:
        dispersion += "-mu"
    if 't' in band_params:
        dispersion += "-2*t*(cos(a*kx)+cos(b*ky))"
    if 'tp' in band_params:
        dispersion += "-4*tp*cos(a*kx)*cos(b*ky)"
    if 'tpp' in band_params:
        dispersion += "-2*tpp*(cos(2*a*kx)+cos(2*b*ky))"
    if 'tz' in band_params:
        dispersion += "-2*tz*(cos(a*kx)-cos(b*ky))**2"
        dispersion += "*cos(a*kx/2)*cos(b*ky/2)*cos(c*kz/2)"
    return dispersion



import unittest
import elecboltz
import numpy as np

class TestGeometricalCalculations(unittest.TestCase):
    def setUp(self):
        # make dummy bandstructure and conductivity objects to inject
        # testing coordinates into them later for just testing the 
        # jacobian calculations
        self.band = elecboltz.BandStructure(
            "Ef * (kx^2 + ky^2 + kz^2)", 1.0, [np.pi, np.pi, np.pi],
            band_params={'Ef': 1.0})
        self.cond = elecboltz.Conductivity(
            self.band, field=[0.0, 0.0, 0.0], scattering_rate=1.0)
    
    def test_regular_octahedron_jacobian(self):
        self.set_up_regular_octahedron()
        self.cond._calculate_jacobian_sums(self.triangle_coordinates)
        neighbor_pairs = {
            (0, 1), (1, 2), (2, 3), (3, 0), (0, 4), (1, 4),
            (2, 4), (3, 4), (0, 5), (1, 5), (2, 5), (3, 5)}
        for i in range(6):
            for j in range(6):
                if i == j:
                    value = 4 * np.sqrt(3)
                elif (i, j) in neighbor_pairs or (j, i) in neighbor_pairs:
                    value = 2 * np.sqrt(3)
                else:
                    value = 0.0
                self.assertAlmostEqual(
                    self.cond._jacobian_sums[i, j], value, 5,
                    f"Regular octahedron Jacobian sum at ({i}, {j})"
                    "is incorrect")

    def test_regular_octahedron_derivative(self):
        self.set_up_regular_octahedron()
        self.cond._calculate_derivative_sums(self.triangle_coordinates)
        derivative_sums = np.zeros((6, 6, 3))
        derivative_sums[0, 1, 2] = -2
        derivative_sums[1, 0, 2] = 2
        derivative_sums[1, 2, 2] = -2
        derivative_sums[2, 1, 2] = 2
        derivative_sums[2, 3, 2] = -2
        derivative_sums[3, 2, 2] = 2
        derivative_sums[3, 0, 2] = -2
        derivative_sums[0, 3, 2] = 2
        derivative_sums[0, 4, 0] = 2
        derivative_sums[4, 0, 0] = -2
        derivative_sums[1, 4, 1] = 2
        derivative_sums[4, 1, 1] = -2
        derivative_sums[2, 4, 0] = -2
        derivative_sums[4, 2, 0] = 2
        derivative_sums[3, 4, 1] = -2
        derivative_sums[4, 3, 1] = 2
        derivative_sums[0, 5, 0] = -2
        derivative_sums[5, 0, 0] = 2
        derivative_sums[1, 5, 1] = -2
        derivative_sums[5, 1, 1] = 2
        derivative_sums[2, 5, 0] = 2
        derivative_sums[5, 2, 0] = -2
        derivative_sums[3, 5, 1] = 2
        derivative_sums[5, 3, 1] = -2
        for i in range(6):
            for j in range(6):
                for k in range(3):
                    self.assertEqual(
                        self.cond._derivatives[k][i, j],
                        derivative_sums[i, j, k],
                        "Regular octahedron derivative sum at"
                        f" ({i}, {j}) is incorrect on axis {k}")

    def set_up_regular_octahedron(self):
        vertices = np.array(
            [[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0],
             [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, -1.0]])
        faces = np.array([
            [0, 1, 4], [0, 5, 1], [1, 2, 4], [1, 5, 2],
            [2, 3, 4], [2, 5, 3], [3, 0, 4], [3, 5, 0]])
        self.set_up_geometry(vertices, faces)

    def set_up_geometry(self, vertices, faces):
        self.band.kpoints = vertices
        self.band.kpoints_periodic = vertices.copy()
        self.band.kfaces = faces
        self.band.kfaces_periodic = faces.copy()
        self.triangle_coordinates = self.band.kpoints[self.band.kfaces]
        self.cond._bandwidth = np.max(np.abs(
            self.band.kfaces - np.roll(self.band.kfaces, 1, axis=1)))


if __name__ == "__main__":
    unittest.main()
